@@ -24,7 +24,8 @@
 #include <ctype.h>
 #include <sys/stat.h>
 #include <limits.h>
-
+#include <openssl/x509.h>  /* work around OpenSSL bug: missing definition of STACK_OF */
+#include <openssl/tls1.h>
 
 const struct vpn_config invalid_cfg = {
 	.gateway_host = {'\0'},
@@ -32,6 +33,9 @@ const struct vpn_config invalid_cfg = {
 	.username = {'\0'},
 	.password = NULL,
 	.otp = {'\0'},
+	.otp_prompt = NULL,
+	.otp_delay = -1,
+	.pinentry = NULL,
 	.realm = {'\0'},
 	.set_routes = -1,
 	.set_dns = -1,
@@ -54,7 +58,10 @@ const struct vpn_config invalid_cfg = {
 	.user_key = NULL,
 	.insecure_ssl = -1,
 	.cipher_list = NULL,
-	.cert_whitelist = NULL
+	.min_tls = -1,
+	.seclevel_1 = -1,
+	.cert_whitelist = NULL,
+	.use_engine = -1
 };
 
 /*
@@ -101,12 +108,44 @@ int strtob(const char *str)
 	else if (isdigit(str[0]) == 0)
 		return -1;
 
-	long int i = strtol(str, NULL, 0);
+	long i = strtol(str, NULL, 0);
 	if (i < 0 || i > 1)
 		return -1;
 	return i;
 }
 
+/*
+ * Converts string to TLS version
+ *
+ * @params[in] str  the string to read from
+ * @return          OpenSSL version or -1
+ */
+int parse_min_tls(const char *str)
+{
+	if (str[0] != '1' || str[1] != '.' || str[2] == 0 || str[3] != 0)
+		return -1;
+	switch (str[2]) {
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+	case '0':
+		return TLS1_VERSION;
+	case '1':
+		return TLS1_1_VERSION;
+	case '2':
+		return TLS1_2_VERSION;
+#ifdef TLS1_3_VERSION
+	/*
+	 * libressl uses version numbers starting with major version 2
+	 * but does not yet support TLS 1.3
+	 */
+	case '3':
+		return TLS1_3_VERSION;
+#endif
+#endif
+	default:
+		return -1;
+	}
+}
+
 /*
  * Reads filename contents and fill cfg with its values.
  *
@@ -190,9 +229,9 @@ int load_config(struct vpn_config *cfg, const char *filename)
 			strncpy(cfg->gateway_host, val, FIELD_SIZE);
 			cfg->gateway_host[FIELD_SIZE] = '\0';
 		} else if (strcmp(key, "port") == 0) {
-			long int port = strtol(val, NULL, 0);
-			if (port <= 0 || port > 65535) {
-				log_warn("Bad port in config file: \"%d\".\n",
+			unsigned long port = strtoul(val, NULL, 0);
+			if (port == 0 || port > 65535) {
+				log_warn("Bad port in config file: \"%lu\".\n",
 				         port);
 				continue;
 			}
@@ -205,6 +244,20 @@ int load_config(struct vpn_config *cfg, const char *filename)
 		} else if (strcmp(key, "otp") == 0) {
 			strncpy(cfg->otp, val, FIELD_SIZE - 1);
 			cfg->otp[FIELD_SIZE] = '\0';
+		} else if (strcmp(key, "otp-prompt") == 0) {
+			free(cfg->otp_prompt);
+			cfg->otp_prompt = strdup(val);
+		} else if (strcmp(key, "otp-delay") == 0) {
+			long otp_delay = strtol(val, NULL, 0);
+			if (otp_delay < 0 || otp_delay > UINT_MAX) {
+				log_warn("Bad value for otp-delay in config file: \"%s\".\n",
+				         val);
+				continue;
+			}
+			cfg->otp_delay = otp_delay;
+		} else if (strcmp(key, "pinentry") == 0) {
+			free(cfg->pinentry);
+			cfg->pinentry = strdup(val);
 		} else if (strcmp(key, "realm") == 0) {
 			strncpy(cfg->realm, val, FIELD_SIZE - 1);
 			cfg->realm[FIELD_SIZE] = '\0';
@@ -233,8 +286,8 @@ int load_config(struct vpn_config *cfg, const char *filename)
 			}
 			cfg->half_internet_routes = half_internet_routes;
 		} else if (strcmp(key, "persistent") == 0) {
-			long int persistent = strtol(val, NULL, 0);
-			if (persistent < 0 || persistent > UINT_MAX) {
+			unsigned long persistent = strtoul(val, NULL, 0);
+			if (persistent > UINT_MAX) {
 				log_warn("Bad value for persistent in config file: \"%s\".\n",
 				         val);
 				continue;
@@ -297,6 +350,8 @@ int load_config(struct vpn_config *cfg, const char *filename)
 		} else if (strcmp(key, "user-cert") == 0) {
 			free(cfg->user_cert);
 			cfg->user_cert = strdup(val);
+			if (strncmp(cfg->user_cert, "pkcs11:", 7) == 0)
+				cfg->use_engine = 1;
 		} else if (strcmp(key, "user-key") == 0) {
 			free(cfg->user_key);
 			cfg->user_key = strdup(val);
@@ -311,6 +366,25 @@ int load_config(struct vpn_config *cfg, const char *filename)
 		} else if (strcmp(key, "cipher-list") == 0) {
 			free(cfg->cipher_list);
 			cfg->cipher_list = strdup(val);
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+		} else if (strcmp(key, "min-tls") == 0) {
+			int min_tls = parse_min_tls(val);
+			if (min_tls == -1) {
+				log_warn("Bad min-tls in config file: \"%s\".\n",
+				         val);
+				continue;
+			} else {
+				cfg->min_tls = min_tls;
+			}
+#endif
+		} else if (strcmp(key, "seclevel-1") == 0) {
+			int seclevel_1 = strtob(val);
+			if (seclevel_1 < 0) {
+				log_warn("Bad seclevel-1 in config file: \"%s\".\n",
+				         val);
+				continue;
+			}
+			cfg->seclevel_1 = seclevel_1;
 		} else {
 			log_warn("Bad key in config file: \"%s\".\n", key);
 			goto err_free;
@@ -330,6 +404,8 @@ err_close:
 void destroy_vpn_config(struct vpn_config *cfg)
 {
 	free(cfg->password);
+	free(cfg->otp_prompt);
+	free(cfg->pinentry);
 #if HAVE_USR_SBIN_PPPD
 	free(cfg->pppd_log);
 	free(cfg->pppd_plugin);
@@ -363,6 +439,12 @@ void merge_config(struct vpn_config *dst, struct vpn_config *src)
 		dst->password = strdup(src->password);
 	if (src->otp[0])
 		strcpy(dst->otp, src->otp);
+	if (src->otp_delay != invalid_cfg.otp_delay)
+		dst->otp_delay = src->otp_delay;
+	if (src->pinentry) {
+		free(dst->pinentry);
+		dst->pinentry = src->pinentry;
+	}
 	if (src->realm[0])
 		strcpy(dst->realm, src->realm);
 	if (src->set_routes != invalid_cfg.set_routes)
@@ -411,6 +493,8 @@ void merge_config(struct vpn_config *dst, struct vpn_config *src)
 	}
 	if (src->user_cert) {
 		free(dst->user_cert);
+		if (strncmp(src->user_cert, "pkcs11:", 7) == 0)
+			dst->use_engine = 1;
 		dst->user_cert = src->user_cert;
 	}
 	if (src->user_key) {
@@ -423,6 +507,10 @@ void merge_config(struct vpn_config *dst, struct vpn_config *src)
 		free(dst->cipher_list);
 		dst->cipher_list = src->cipher_list;
 	}
+	if (src->min_tls > 0)
+		dst->min_tls = src->min_tls;
+	if (src->seclevel_1 != invalid_cfg.seclevel_1)
+		dst->seclevel_1 = src->seclevel_1;
 	if (src->cert_whitelist) {
 		while (dst->cert_whitelist != NULL) {
 			struct x509_digest *tmp = dst->cert_whitelist->next;
