@@ -29,16 +29,17 @@
 
 #define PWD_BUFSIZ	4096
 
-
 #if HAVE_USR_SBIN_PPPD
 #define PPPD_USAGE \
-"                    [--pppd-no-peerdns] [--pppd-log=<file>]\n" \
+"                    [--pppd-use-peerdns=<0|1>] [--pppd-log=<file>]\n" \
 "                    [--pppd-ifname=<string>] [--pppd-ipparam=<string>]\n" \
 "                    [--pppd-call=<name>] [--pppd-plugin=<file>]\n"
 
 #define PPPD_HELP \
-"  --pppd-no-peerdns             Do not ask peer ppp server for DNS server addresses\n" \
-"                                and do not make pppd rewrite /etc/resolv.conf\n" \
+"  --pppd-use-peerdns=[01]       Whether to ask peer ppp server for DNS server\n" \
+"                                addresses and make pppd rewrite /etc/resolv.conf.\n" \
+"  --pppd-no-peerdns             Same as --pppd-use-peerdns=0. pppd will not\n" \
+"                                modify DNS resolution then.\n" \
 "  --pppd-log=<file>             Set pppd in debug mode and save its logs into\n" \
 "                                <file>.\n" \
 "  --pppd-plugin=<file>          Use specified pppd plugin instead of configuring\n" \
@@ -60,7 +61,9 @@
 
 #define usage \
 "Usage: openfortivpn [<host>[:<port>]] [-u <user>] [-p <pass>]\n" \
-"                    [--realm=<realm>] [--otp=<otp>] [--set-routes=<0|1>]\n" \
+"                    [--pinentry=<program>]\n" \
+"                    [--realm=<realm>] [--otp=<otp>] [--otp-delay=<delay>]\n" \
+"                    [--otp-prompt=<prompt>] [--set-routes=<0|1>]\n" \
 "                    [--half-internet-routes=<0|1>] [--set-dns=<0|1>]\n" \
 PPPD_USAGE \
 "                    [--ca-file=<file>]\n" \
@@ -78,6 +81,13 @@ PPPD_USAGE \
 "the gateway and this process.\n" \
 "\n"
 
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+#define help_min_tls ""
+#else
+#define help_min_tls " This option\n" \
+"                                is not supported by your OpenSSL library."
+#endif
+
 
 #define help_options \
 "Options:\n" \
@@ -88,21 +98,25 @@ PPPD_USAGE \
 "  -u <user>, --username=<user>  VPN account username.\n" \
 "  -p <pass>, --password=<pass>  VPN account password.\n" \
 "  -o <otp>, --otp=<otp>         One-Time-Password.\n" \
-"  --realm=<realm>               Use specified authentication realm on VPN gateway\n" \
+"  --otp-prompt=<prompt>         Search for the OTP prompt starting with this string\n" \
+"  --otp-delay=<delay>           Wait <delay> seconds before sending the OTP.\n" \
+"  --pinentry=<program>          Use the program to supply a secret instead of asking for it\n" \
+"  --realm=<realm>               Use specified authentication realm.\n" \
+"  --set-routes=[01]             Set if openfortivpn should configure routes\n" \
 "                                when tunnel is up.\n" \
-"  --set-routes=[01]             Set if openfortivpn should configure output routes through\n" \
-"                                the VPN when tunnel is up.\n" \
 "  --no-routes                   Do not configure routes, same as --set-routes=0.\n" \
 "  --half-internet-routes=[01]   Add two 0.0.0.0/1 and 128.0.0.0/1 routes with higher\n" \
 "                                priority instead of replacing the default route.\n" \
-"  --set-dns=[01]                Set if openfortivpn should add VPN name servers in\n" \
-"                                /etc/resolv.conf\n" \
+"  --set-dns=[01]                Set if openfortivpn should add DNS name servers\n" \
+"                                and domain search list in /etc/resolv.conf.\n" \
+"                                If installed resolvconf is used for the update.\n" \
 "  --no-dns                      Do not reconfigure DNS, same as --set-dns=0\n" \
 "  --ca-file=<file>              Use specified PEM-encoded certificate bundle\n" \
 "                                instead of system-wide store to verify the gateway\n" \
 "                                certificate.\n" \
 "  --user-cert=<file>            Use specified PEM-encoded certificate if the server\n" \
 "                                requires authentication with a certificate.\n" \
+"  --user-cert=pkcs11:           Use smartcard. Takes also partial or full PKCS11-URI.\n" \
 "  --user-key=<file>             Use specified PEM-encoded key if the server requires\n" \
 "                                authentication with a certificate.\n" \
 "  --use-syslog                  Log to syslog instead of terminal.\n" \
@@ -113,23 +127,28 @@ PPPD_USAGE \
 "                                This option can be used multiple times to trust\n" \
 "                                several certificates.\n" \
 "  --insecure-ssl                Do not disable insecure SSL protocols/ciphers.\n" \
-"                                If your server requires a specific cipher, consider\n" \
-"                                using --cipher-list instead.\n" \
-"  --cipher-list=<ciphers>       Openssl ciphers to use. If default does not work\n" \
+"                                Also enable TLS v1.0 if applicable.\n" \
+"                                If your server requires a specific cipher or protocol,\n" \
+"                                consider using --cipher-list and/or --min-tls instead.\n" \
+"  --cipher-list=<ciphers>       OpenSSL ciphers to use. If default does not work\n" \
 "                                you can try with the cipher suggested in the output\n" \
 "                                of 'openssl s_client -connect <host:port>'\n" \
 "                                (e.g. AES256-GCM-SHA384)\n" \
-PPPD_HELP \
+"  --min-tls                     Use minimum TLS version instead of system default.\n" \
+"                                Valid values are 1.0, 1.1, 1.2, 1.3." help_min_tls "\n" \
+"  --seclevel-1                  If --cipher-list is not specified, add @SECLEVEL=1 to\n" \
+"                                (compiled in) list of ciphers. This lowers limits on\n" \
+"                                dh key.\n" \
 "  --persistent=<interval>       Run the vpn persistently in a loop and try to re-\n" \
 "                                connect every <interval> seconds when dropping out\n" \
 "  -v                            Increase verbosity. Can be used multiple times\n" \
 "                                to be even more verbose.\n" \
 "  -q                            Decrease verbosity. Can be used multiple times\n" \
-"                                to be even less verbose.\n" \
-"\n"
+"                                to be even less verbose.\n"
 
 
 #define help_config \
+"\n" \
 "Config file:\n" \
 "  Options can be taken from a configuration file. Options passed in the\n" \
 "  command line will override those from the config file, though. The default\n" \
@@ -158,6 +177,9 @@ int main(int argc, char **argv)
 		.username = {'\0'},
 		.password = NULL,
 		.otp = {'\0'},
+		.otp_prompt = NULL,
+		.otp_delay = 0,
+		.pinentry = NULL,
 		.realm = {'\0'},
 		.set_routes = 1,
 		.set_dns = 1,
@@ -165,7 +187,7 @@ int main(int argc, char **argv)
 		.half_internet_routes = 0,
 		.persistent = 0,
 #if HAVE_USR_SBIN_PPPD
-		.pppd_use_peerdns = 1,
+		.pppd_use_peerdns = 0,
 		.pppd_log = NULL,
 		.pppd_plugin = NULL,
 		.pppd_ipparam = NULL,
@@ -179,6 +201,8 @@ int main(int argc, char **argv)
 		.user_cert = NULL,
 		.user_key = NULL,
 		.insecure_ssl = 0,
+		.min_tls = 0,
+		.seclevel_1 = 0,
 		.cipher_list = NULL,
 		.cert_whitelist = NULL
 	};
@@ -188,10 +212,13 @@ int main(int argc, char **argv)
 		{"help",            no_argument,       0, 'h'},
 		{"version",         no_argument,       0, 0},
 		{"config",          required_argument, 0, 'c'},
+		{"pinentry",        required_argument, 0, 0},
 		{"realm",           required_argument, 0, 0},
 		{"username",        required_argument, 0, 'u'},
 		{"password",        required_argument, 0, 'p'},
 		{"otp",             required_argument, 0, 'o'},
+		{"otp-prompt",      required_argument, 0, 0},
+		{"otp-delay",       required_argument, 0, 0},
 		{"set-routes",	    required_argument, 0, 0},
 		{"no-routes",       no_argument, &cli_cfg.set_routes, 0},
 		{"half-internet-routes", required_argument, 0, 0},
@@ -205,7 +232,12 @@ int main(int argc, char **argv)
 		{"trusted-cert",    required_argument, 0, 0},
 		{"insecure-ssl",    no_argument, &cli_cfg.insecure_ssl, 1},
 		{"cipher-list",     required_argument, 0, 0},
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+		{"min-tls",         required_argument, 0, 0},
+#endif
+		{"seclevel-1",      no_argument, &cli_cfg.seclevel_1, 1},
 #if HAVE_USR_SBIN_PPPD
+		{"pppd-use-peerdns", required_argument, 0, 0},
 		{"pppd-no-peerdns", no_argument, &cli_cfg.pppd_use_peerdns, 0},
 		{"pppd-log",        required_argument, 0, 0},
 		{"pppd-plugin",     required_argument, 0, 0},
@@ -245,6 +277,17 @@ int main(int argc, char **argv)
 				goto exit;
 			}
 #if HAVE_USR_SBIN_PPPD
+			if (strcmp(long_options[option_index].name,
+			           "pppd-use-peerdns") == 0) {
+				int pppd_use_peerdns = strtob(optarg);
+				if (pppd_use_peerdns < 0) {
+					log_warn("Bad pppd-use-peerdns option: \"%s\"\n",
+					         optarg);
+					break;
+				}
+				cli_cfg.pppd_use_peerdns = pppd_use_peerdns;
+				break;
+			}
 			if (strcmp(long_options[option_index].name,
 			           "pppd-log") == 0) {
 				cli_cfg.pppd_log = strdup(optarg);
@@ -300,6 +343,11 @@ int main(int argc, char **argv)
 				cli_cfg.user_key = strdup(optarg);
 				break;
 			}
+			if (strcmp(long_options[option_index].name,
+			           "pinentry") == 0) {
+				cli_cfg.pinentry = strdup(optarg);
+				break;
+			}
 			if (strcmp(long_options[option_index].name,
 			           "realm") == 0) {
 				strncpy(cli_cfg.realm, optarg, FIELD_SIZE);
@@ -317,6 +365,24 @@ int main(int argc, char **argv)
 				cli_cfg.cipher_list = strdup(optarg);
 				break;
 			}
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+			if (strcmp(long_options[option_index].name,
+			           "min-tls") == 0) {
+				int min_tls = parse_min_tls(optarg);
+				if (min_tls == -1) {
+					log_warn("Bad min-tls option: \"%s\"\n",
+					         optarg);
+				} else {
+					cli_cfg.min_tls = min_tls;
+				}
+				break;
+			}
+#endif
+			if (strcmp(long_options[option_index].name,
+			           "otp-prompt") == 0) {
+				cli_cfg.otp_prompt = strdup(optarg);
+				break;
+			}
 			if (strcmp(long_options[option_index].name,
 			           "set-routes") == 0) {
 				int set_routes = strtob(optarg);
@@ -339,9 +405,20 @@ int main(int argc, char **argv)
 				cli_cfg.half_internet_routes = half_internet_routes;
 				break;
 			}
+			if (strcmp(long_options[option_index].name,
+			           "otp-delay") == 0) {
+				long otp_delay = strtol(optarg, NULL, 0);
+				if (otp_delay < 0 || otp_delay > UINT_MAX) {
+					log_warn("Bad otp-delay option: \"%s\"\n",
+					         optarg);
+					break;
+				}
+				cli_cfg.otp_delay = otp_delay;
+				break;
+			}
 			if (strcmp(long_options[option_index].name,
 			           "persistent") == 0) {
-				long int persistent = strtol(optarg, NULL, 0);
+				long persistent = strtol(optarg, NULL, 0);
 				if (persistent < 0 || persistent > UINT_MAX) {
 					log_warn("Bad persistent option: \"%s\"\n",
 					         optarg);
@@ -362,7 +439,8 @@ int main(int argc, char **argv)
 			}
 			goto user_error;
 		case 'h':
-			printf("%s%s%s%s", usage, summary, help_options, help_config);
+			printf("%s%s%s%s%s", usage, summary, help_options,
+			       PPPD_HELP, help_config);
 			ret = EXIT_SUCCESS;
 			goto exit;
 		case 'v':
@@ -396,7 +474,9 @@ int main(int argc, char **argv)
 	if (cli_cfg.password != NULL && cli_cfg.password[0] != '\0')
 		log_warn("You should not pass the password on the command line. Type it interactively or use a config file instead.\n");
 
-	log_debug("openfortivpn " VERSION "\n", config_file);
+	log_debug_all("ATTENTION: the output contains sensitive information such as the THE CLEAR TEXT PASSWORD.\n");
+
+	log_debug("openfortivpn " VERSION "\n");
 
 	// Load config file
 	if (config_file[0] != '\0') {
@@ -407,6 +487,14 @@ int main(int argc, char **argv)
 			log_warn("Could not load config file \"%s\" (%s).\n",
 			         config_file, err_cfg_str(ret));
 	}
+	if (cfg.password != NULL && cli_cfg.password == NULL) {
+		if (cfg.password[0] == '\0')
+			log_debug("Disabled password due to empty entry in config file \"%s\"\n",
+			          config_file);
+		else
+			log_debug("Loaded password from config file \"%s\"\n",
+			          config_file);
+	}
 	// Then apply CLI config
 	merge_config(&cfg, &cli_cfg);
 	set_syslog(cfg.use_syslog);
@@ -419,7 +507,7 @@ int main(int argc, char **argv)
 			port_str[0] = '\0';
 			port_str++;
 			cfg.gateway_port = strtol(port_str, NULL, 0);
-			if (cfg.gateway_port <= 0 || cfg.gateway_port > 65535) {
+			if (cfg.gateway_port == 0 || cfg.gateway_port > 65535) {
 				log_error("Specify a valid port.\n");
 				goto user_error;
 			}
@@ -434,29 +522,27 @@ int main(int argc, char **argv)
 		goto user_error;
 	}
 	// Check username
-	if (cfg.username[0] == '\0') {
-		log_error("Specify an username.\n");
+	if (cfg.username[0] == '\0' && cfg.user_cert == NULL) {
+		log_error("Specify a username.\n");
 		goto user_error;
 	}
-	// If no password given, interactively ask user
-	if (cfg.password == NULL || cfg.password[0] == '\0') {
-		free(cfg.password);
+	// If username but no password given, interactively ask user
+	if (cfg.password == NULL && cfg.username[0] != '\0') {
 		char *tmp_password = malloc(PWD_BUFSIZ); // allocate large buffer
-		read_password("VPN account password: ", tmp_password, PWD_BUFSIZ);
+		read_password(cfg.pinentry, "password",
+		              "VPN account password: ", tmp_password, PWD_BUFSIZ);
 		cfg.password = strdup(tmp_password); // copy string of correct size
 		free(tmp_password);
 	}
-	// Check password
-	if (cfg.password[0] == '\0') {
-		log_error("Specify a password.\n");
-		goto user_error;
-	}
-
 	log_debug("Config host = \"%s\"\n", cfg.gateway_host);
 	log_debug("Config realm = \"%s\"\n", cfg.realm);
 	log_debug("Config port = \"%d\"\n", cfg.gateway_port);
-	log_debug("Config username = \"%s\"\n", cfg.username);
-	log_debug("Config password = \"%s\"\n", "********");
+	if (cfg.username[0] != '\0')
+		log_debug("Config username = \"%s\"\n", cfg.username);
+	if (cfg.password != NULL)
+		log_debug_all("Config password = \"%s\"\n", cfg.password);
+	else
+		cfg.password = strdup("");
 	if (cfg.otp[0] != '\0')
 		log_debug("One-time password = \"%s\"\n", cfg.otp);
 
