@@ -29,6 +29,9 @@
 #include "tunnel.h"
 #include "http.h"
 #include "log.h"
+#ifndef HAVE_X509_CHECK_HOST
+#include "openssl_hostname_validation.h"
+#endif
 
 #include <unistd.h>
 #include <fcntl.h>
@@ -42,6 +45,7 @@
 #include <sys/ioctl.h>
 #include <openssl/err.h>
 #include <openssl/x509v3.h>
+#include <openssl/engine.h>
 #if HAVE_PTY_H
 #include <pty.h>
 #elif HAVE_UTIL_H
@@ -50,36 +54,50 @@
 #include <termios.h>
 #include <signal.h>
 #include <sys/wait.h>
-#include <assert.h>
 #if HAVE_SYSTEMD
 #include <systemd/sd-daemon.h>
 #endif
 
+// we use this constant in the source, so define a fallback if not defined
+#ifndef OPENSSL_API_COMPAT
+#define OPENSSL_API_COMPAT 0x0908000L
+#endif
+
 struct ofv_varr {
-	unsigned cap;		// current capacity
-	unsigned off;		// next slot to write, always < max(cap - 1, 1)
+	unsigned int cap;	// current capacity
+	unsigned int off;	// next slot to write, always < max(cap - 1, 1)
 	const char **data;	// NULL terminated
 };
 
-static void ofv_append_varr(struct ofv_varr *p, const char *x)
+static int ofv_append_varr(struct ofv_varr *p, const char *x)
 {
 	if (p->off + 1 >= p->cap) {
 		const char **ndata;
-		unsigned ncap = (p->off + 1) * 2;
-		assert(p->off + 1 < ncap);
+		unsigned int ncap = (p->off + 1) * 2;
+		if (p->off + 1 >= ncap) {
+			log_error("%s: ncap exceeded\n", __func__);
+			return 1;
+		};
 		ndata = realloc(p->data, ncap * sizeof(const char *));
 		if (ndata) {
 			p->data = ndata;
 			p->cap = ncap;
 		} else {
 			log_error("realloc: %s\n", strerror(errno));
-			assert(ndata);
-			return;
+			return 1;
 		}
 	}
-	assert(p->off + 1 < p->cap);
+	if (p->data == NULL) {
+		log_error("%s: NULL data\n", __func__);
+		return 1;
+	}
+	if (p->off + 1 >= p->cap) {
+		log_error("%s: cap exceeded in p\n", __func__);
+		return 1;
+	}
 	p->data[p->off] = x;
 	p->data[++p->off] = NULL;
+	return 0;
 }
 
 static int on_ppp_if_up(struct tunnel *tunnel)
@@ -93,9 +111,8 @@ static int on_ppp_if_up(struct tunnel *tunnel)
 
 		ret = ipv4_set_tunnel_routes(tunnel);
 
-		if (ret != 0) {
+		if (ret != 0)
 			log_warn("Adding route table is incomplete. Please check route table.\n");
-		}
 	}
 
 	if (tunnel->config->set_dns) {
@@ -133,6 +150,8 @@ static int pppd_run(struct tunnel *tunnel)
 {
 	pid_t pid;
 	int amaster;
+	int slave_stderr;
+
 #ifdef HAVE_STRUCT_TERMIOS
 	struct termios termp = {
 		.c_cflag = B9600,
@@ -148,42 +167,53 @@ static int pppd_run(struct tunnel *tunnel)
 	}
 	log_debug("ppp_path: %s\n", ppp_path);
 
+	slave_stderr = dup(STDERR_FILENO);
+
+	if (slave_stderr < 0) {
+		log_error("slave stderr %s\n", strerror(errno));
+		return 1;
+	}
+
 #ifdef HAVE_STRUCT_TERMIOS
 	pid = forkpty(&amaster, NULL, &termp, NULL);
 #else
 	pid = forkpty(&amaster, NULL, NULL, NULL);
 #endif
 
-	if (pid == -1) {
-		log_error("forkpty: %s\n", strerror(errno));
-		return 1;
-	} else if (pid == 0) { // child process
+	if (pid == 0) { // child process
 
 		struct ofv_varr pppd_args = { 0, 0, NULL };
 
+		dup2(slave_stderr, STDERR_FILENO);
+		close(slave_stderr);
+
 #if HAVE_USR_SBIN_PPP
 		/*
-		* assume there is a default configuration to start.
-		* Support for taking options from the command line
-		* e.g. the name of the configuration or options
-		* to send interactively to ppp will be added later
-		*/
-		const char *v[] = {
+		 * assume there is a default configuration to start.
+		 * Support for taking options from the command line
+		 * e.g. the name of the configuration or options
+		 * to send interactively to ppp will be added later
+		 */
+		static const char *const v[] = {
 			ppp_path,
 			"-direct"
 		};
-		for (unsigned i = 0; i < sizeof v/sizeof v[0]; i++)
-			ofv_append_varr(&pppd_args, v[i]);
+		for (unsigned int i = 0; i < ARRAY_SIZE(v); i++)
+			if (ofv_append_varr(&pppd_args, v[i]))
+				return 1;
 #endif
 #if HAVE_USR_SBIN_PPPD
 		if (tunnel->config->pppd_call) {
-			ofv_append_varr(&pppd_args, ppp_path);
-			ofv_append_varr(&pppd_args, "call");
-			ofv_append_varr(&pppd_args, tunnel->config->pppd_call);
+			if (ofv_append_varr(&pppd_args, ppp_path))
+				return 1;
+			if (ofv_append_varr(&pppd_args, "call"))
+				return 1;
+			if (ofv_append_varr(&pppd_args, tunnel->config->pppd_call))
+				return 1;
 		} else {
-			const char *v[] = {
+			static const char *const v[] = {
 				ppp_path,
-				"38400", // speed
+				"115200", // speed
 				":192.0.2.1", // <local_IP_address>:<remote_IP_address>
 				"noipdefault",
 				"noaccomp",
@@ -196,32 +226,53 @@ static int pppd_run(struct tunnel *tunnel)
 				"lcp-max-configure", "40",
 				"mru", "1354"
 			};
-			for (unsigned i = 0; i < ARRAY_SIZE(v); i++)
-				ofv_append_varr(&pppd_args, v[i]);
+			for (unsigned int i = 0; i < ARRAY_SIZE(v); i++)
+				if (ofv_append_varr(&pppd_args, v[i]))
+					return 1;
 		}
 		if (tunnel->config->pppd_use_peerdns)
-			ofv_append_varr(&pppd_args, "usepeerdns");
+			if (ofv_append_varr(&pppd_args, "usepeerdns"))
+				return 1;
 		if (tunnel->config->pppd_log) {
-			ofv_append_varr(&pppd_args, "debug");
-			ofv_append_varr(&pppd_args, "logfile");
-			ofv_append_varr(&pppd_args, tunnel->config->pppd_log);
+			if (ofv_append_varr(&pppd_args, "debug"))
+				return 1;
+			if (ofv_append_varr(&pppd_args, "logfile"))
+				return 1;
+			if (ofv_append_varr(&pppd_args, tunnel->config->pppd_log))
+				return 1;
+		} else {
+			/*
+			 * pppd defaults to logging to fd=1, clobbering the
+			 * actual PPP data
+			 */
+			if (ofv_append_varr(&pppd_args, "logfd"))
+				return 1;
+			if (ofv_append_varr(&pppd_args, "2"))
+				return 1;
 		}
 		if (tunnel->config->pppd_plugin) {
-			ofv_append_varr(&pppd_args, "plugin");
-			ofv_append_varr(&pppd_args, tunnel->config->pppd_plugin);
+			if (ofv_append_varr(&pppd_args, "plugin"))
+				return 1;
+			if (ofv_append_varr(&pppd_args, tunnel->config->pppd_plugin))
+				return 1;
 		}
 		if (tunnel->config->pppd_ipparam) {
-			ofv_append_varr(&pppd_args, "ipparam");
-			ofv_append_varr(&pppd_args, tunnel->config->pppd_ipparam);
+			if (ofv_append_varr(&pppd_args, "ipparam"))
+				return 1;
+			if (ofv_append_varr(&pppd_args, tunnel->config->pppd_ipparam))
+				return 1;
 		}
 		if (tunnel->config->pppd_ifname) {
-			ofv_append_varr(&pppd_args, "ifname");
-			ofv_append_varr(&pppd_args, tunnel->config->pppd_ifname);
+			if (ofv_append_varr(&pppd_args, "ifname"))
+				return 1;
+			if (ofv_append_varr(&pppd_args, tunnel->config->pppd_ifname))
+				return 1;
 		}
 #endif
 #if HAVE_USR_SBIN_PPP
 		if (tunnel->config->ppp_system) {
-			ofv_append_varr(&pppd_args, tunnel->config->ppp_system);
+			if (ofv_append_varr(&pppd_args, tunnel->config->ppp_system))
+				return 1;
 		}
 #endif
 
@@ -229,13 +280,14 @@ static int pppd_run(struct tunnel *tunnel)
 		execv(pppd_args.data[0], (char *const *)pppd_args.data);
 		free(pppd_args.data);
 
-		/*
-		 * The following call to fprintf() doesn't work, probably
-		 * because of the prior call to forkpty().
-		 * TODO: print a meaningful message using strerror(errno)
-		 */
 		fprintf(stderr, "execvp: %s\n", strerror(errno));
 		_exit(EXIT_FAILURE);
+	} else {
+		close(slave_stderr);
+		if (pid == -1) {
+			log_error("forkpty: %s\n", strerror(errno));
+			return 1;
+		}
 	}
 
 	// Set non-blocking
@@ -357,12 +409,12 @@ int ppp_interface_is_up(struct tunnel *tunnel)
 		            (tunnel->config->pppd_ifname
 		             && strstr(ifa->ifa_name, tunnel->config->pppd_ifname)
 		             != NULL)
-		            || strstr(ifa->ifa_name, "ppp") != NULL)
+		            || strstr(ifa->ifa_name, "ppp") != NULL
 #endif
 #if HAVE_USR_SBIN_PPP
-		    strstr(ifa->ifa_name, "tun") != NULL)
+		            strstr(ifa->ifa_name, "tun") != NULL
 #endif
-			&& ifa->ifa_flags& IFF_UP) {
+		    ) && ifa->ifa_flags & IFF_UP) {
 			if (&(ifa->ifa_addr->sa_family) != NULL
 			    && ifa->ifa_addr->sa_family == AF_INET) {
 				struct in_addr if_ip_addr =
@@ -452,7 +504,7 @@ static int tcp_connect(struct tunnel *tunnel)
 		if (proxy_port != NULL) {
 			proxy_port[0] = '\0';
 			proxy_port++;
-			server.sin_port = htons(strtol(proxy_port, NULL, 10));
+			server.sin_port = htons(strtoul(proxy_port, NULL, 10));
 		} else {
 			server.sin_port = htons(tunnel->config->gateway_port);
 		}
@@ -462,7 +514,7 @@ static int tcp_connect(struct tunnel *tunnel)
 		log_debug("proxy_host: %s\n", proxy_host);
 		log_debug("proxy_port: %s\n", proxy_port);
 		server.sin_addr.s_addr = inet_addr(proxy_host);
-		// if host is given as fqhn we have to do a dns lookup
+		// if host is given as a FQDN we have to do a DNS lookup
 		if (server.sin_addr.s_addr == INADDR_NONE) {
 			const struct addrinfo hints = { .ai_family = AF_INET };
 			struct addrinfo *result = NULL;
@@ -532,13 +584,13 @@ static int tcp_connect(struct tunnel *tunnel)
 			 */
 			ssize_t bytes_read = read(handle, &(request[j]), 1);
 			if (bytes_read < 1) {
-				log_error("Proxy response is unexpectedly large and cannot fit in the %d-bytes buffer.\n",
+				log_error("Proxy response is unexpectedly large and cannot fit in the %lu-bytes buffer.\n",
 				          ARRAY_SIZE(request));
 				goto err_proxy_response;
 			}
 
 			// detect "200"
-			const char HTTP_STATUS_200[] = "200";
+			static const char HTTP_STATUS_200[] = "200";
 			response = strstr(request, HTTP_STATUS_200);
 
 			// detect end-of-line after "200"
@@ -560,7 +612,7 @@ static int tcp_connect(struct tunnel *tunnel)
 				 * 	recognize a single LF as a line terminator
 				 * 	and ignore the leading CR.
 				 */
-				static const char *HTTP_EOL[] = {
+				static const char *const HTTP_EOL[] = {
 					"\r\n\r\n",
 					"\n\n"
 				};
@@ -603,7 +655,6 @@ static int ssl_verify_cert(struct tunnel *tunnel)
 	char *line;
 	int i;
 	X509_NAME *subj;
-	char common_name[FIELD_SIZE + 1];
 
 	SSL_set_verify(tunnel->ssl_handle, SSL_VERIFY_PEER, NULL);
 
@@ -617,16 +668,15 @@ static int ssl_verify_cert(struct tunnel *tunnel)
 
 #ifdef HAVE_X509_CHECK_HOST
 	// Use OpenSSL native host validation if v >= 1.0.2.
-	if (X509_check_host(cert, common_name, FIELD_SIZE, 0, NULL))
+	// compare against gateway_host and correctly check return value
+	// to fix piror Incorrect use of X509_check_host
+	if (X509_check_host(cert, tunnel->config->gateway_host,
+	                    0, 0, NULL) == 1)
 		cert_valid = 1;
 #else
-	// Use explicit Common Name check if native validation not available.
-	// Note: this will ignore Subject Alternative Name fields.
-	if (subj
-	    && X509_NAME_get_text_by_NID(subj, NID_commonName, common_name,
-	                                 FIELD_SIZE) > 0
-	    && strncasecmp(common_name, tunnel->config->gateway_host,
-	                   FIELD_SIZE) == 0)
+	// Use validate_hostname form iSECPartners if native validation not available
+	// in order to avoid TLS Certificate CommonName NULL Byte Vulnerability
+	if (validate_hostname(tunnel->config->gateway_host, cert) == MatchFound)
 		cert_valid = 1;
 #endif
 
@@ -702,7 +752,7 @@ static void ssl_disconnect(struct tunnel *tunnel)
 }
 
 /*
- * Connects to the gateway and initiate a SSL session.
+ * Connects to the gateway and initiate an SSL session.
  */
 int ssl_connect(struct tunnel *tunnel)
 {
@@ -712,10 +762,13 @@ int ssl_connect(struct tunnel *tunnel)
 	if (tunnel->ssl_socket == -1)
 		return 1;
 
+	// registration is deprecated from OpenSSL 1.1.0 onward
+#if OPENSSL_API_COMPAT < 0x10100000L
 	// Register the error strings for libcrypto & libssl
 	SSL_load_error_strings();
 	// Register the available ciphers and digests
 	SSL_library_init();
+#endif
 
 	tunnel->ssl_context = SSL_CTX_new(SSLv23_client_method());
 	if (tunnel->ssl_context == NULL) {
@@ -738,32 +791,94 @@ int ssl_connect(struct tunnel *tunnel)
 		}
 	}
 
-	if (tunnel->config->user_cert) {
-		if (!SSL_CTX_use_certificate_file(
-		            tunnel->ssl_context, tunnel->config->user_cert,
-		            SSL_FILETYPE_PEM)) {
-			log_error("SSL_CTX_use_certificate_file: %s\n",
+	/* Use engine for PIV if user-cert config starts with pkcs11 URI: */
+	if (tunnel->config->use_engine > 0) {
+
+		ENGINE *e;
+		ENGINE_load_builtin_engines();
+		e = ENGINE_by_id("pkcs11");
+		if (!e) {
+			log_error("Could not load pkcs11 Engine: %s\n",
 			          ERR_error_string(ERR_peek_last_error(), NULL));
 			return 1;
 		}
-	}
+		if (!ENGINE_init(e)) {
+			log_error("Could not init pkcs11 Engine: %s\n",
+			          ERR_error_string(ERR_peek_last_error(), NULL));
+			ENGINE_free(e);
+			return 1;
+		}
+		if (!ENGINE_set_default_RSA(e))
+			abort();
 
-	if (tunnel->config->user_key) {
-		if (!SSL_CTX_use_PrivateKey_file(
-		            tunnel->ssl_context, tunnel->config->user_key,
-		            SSL_FILETYPE_PEM)) {
-			log_error("SSL_CTX_use_PrivateKey_file: %s\n",
+		ENGINE_finish(e);
+		ENGINE_free(e);
+
+		struct token parms;
+		parms.uri = tunnel->config->user_cert;
+		parms.cert = NULL;
+
+		if (!ENGINE_ctrl_cmd(e, "LOAD_CERT_CTRL", 0, &parms, NULL, 1)) {
+			log_error("PKCS11 ENGINE_ctrl_cmd: %s\n",
+			          ERR_error_string(ERR_peek_last_error(), NULL));
+			return 1;
+		}
+
+		if (!SSL_CTX_use_certificate(tunnel->ssl_context, parms.cert)) {
+			log_error("PKCS11 SSL_CTX_use_certificate: %s\n",
+			          ERR_error_string(ERR_peek_last_error(), NULL));
+			return 1;
+		}
+
+		EVP_PKEY * privkey = ENGINE_load_private_key(
+		                             e, parms.uri, UI_OpenSSL(), NULL);
+		if (!privkey) {
+			log_error("PKCS11 ENGINE_load_private_key: %s\n",
+			          ERR_error_string(ERR_peek_last_error(), NULL));
+			return 1;
+		}
+
+		if (!SSL_CTX_use_PrivateKey(tunnel->ssl_context, privkey)) {
+			log_error("PKCS11 SSL_CTX_use_PrivateKey_file: %s\n",
 			          ERR_error_string(ERR_peek_last_error(), NULL));
 			return 1;
 		}
-	}
 
-	if (tunnel->config->user_cert && tunnel->config->user_key) {
 		if (!SSL_CTX_check_private_key(tunnel->ssl_context)) {
-			log_error("SSL_CTX_check_private_key: %s\n",
+			log_error("PKCS11 SSL_CTX_check_private_key: %s\n",
 			          ERR_error_string(ERR_peek_last_error(), NULL));
 			return 1;
 		}
+
+	} else {        /* end PKCS11-engine */
+
+		if (tunnel->config->user_cert) {
+			if (!SSL_CTX_use_certificate_file(
+			            tunnel->ssl_context, tunnel->config->user_cert,
+			            SSL_FILETYPE_PEM)) {
+				log_error("SSL_CTX_use_certificate_file: %s\n",
+				          ERR_error_string(ERR_peek_last_error(), NULL));
+				return 1;
+			}
+		}
+
+		if (tunnel->config->user_key) {
+			if (!SSL_CTX_use_PrivateKey_file(
+			            tunnel->ssl_context, tunnel->config->user_key,
+			            SSL_FILETYPE_PEM)) {
+				log_error("SSL_CTX_use_PrivateKey_file: %s\n",
+				          ERR_error_string(ERR_peek_last_error(), NULL));
+				return 1;
+			}
+		}
+
+		if (tunnel->config->user_cert && tunnel->config->user_key) {
+			if (!SSL_CTX_check_private_key(tunnel->ssl_context)) {
+				log_error("SSL_CTX_check_private_key: %s\n",
+				          ERR_error_string(ERR_peek_last_error(), NULL));
+				return 1;
+			}
+		}
 	}
 
 	if (!tunnel->config->insecure_ssl) {
@@ -785,13 +900,28 @@ int ssl_connect(struct tunnel *tunnel)
 		return 1;
 	}
 
-	if (!tunnel->config->insecure_ssl && !tunnel->config->cipher_list) {
-		const char *cipher_list = "HIGH:!aNULL:!kRSA:!PSK:!SRP:!MD5:!RC4";
-
-		tunnel->config->cipher_list = strdup(cipher_list);
+	if (!tunnel->config->insecure_ssl) {
+		if (!tunnel->config->cipher_list) {
+			const char *cipher_list;
+			if (tunnel->config->seclevel_1)
+				cipher_list = "HIGH:!aNULL:!kRSA:!PSK:!SRP:!MD5:!RC4@SECLEVEL=1";
+			else
+				cipher_list = "HIGH:!aNULL:!kRSA:!PSK:!SRP:!MD5:!RC4";
+			tunnel->config->cipher_list = strdup(cipher_list);
+		}
+	} else {
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+		if (tunnel->config->min_tls <= 0)
+			tunnel->config->min_tls = TLS1_VERSION;
+#endif
+		if (!tunnel->config->cipher_list && tunnel->config->seclevel_1) {
+			const char *cipher_list = "DEFAULT@SECLEVEL=1";
+			tunnel->config->cipher_list = strdup(cipher_list);
+		}
 	}
 
 	if (tunnel->config->cipher_list) {
+		log_debug("Setting cipher list to: %s\n", tunnel->config->cipher_list);
 		if (!SSL_set_cipher_list(tunnel->ssl_handle,
 		                         tunnel->config->cipher_list)) {
 			log_error("SSL_set_cipher_list failed: %s\n",
@@ -800,6 +930,19 @@ int ssl_connect(struct tunnel *tunnel)
 		}
 	}
 
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+	if (tunnel->config->min_tls > 0) {
+		log_debug("Setting min proto version to: 0x%x\n",
+		          tunnel->config->min_tls);
+		if (!SSL_set_min_proto_version(tunnel->ssl_handle,
+		                               tunnel->config->min_tls)) {
+			log_error("SSL_set_min_proto_version failed: %s\n",
+			          ERR_error_string(ERR_peek_last_error(), NULL));
+			return 1;
+		}
+	}
+#endif
+
 	if (!SSL_set_fd(tunnel->ssl_handle, tunnel->ssl_socket)) {
 		log_error("SSL_set_fd: %s\n",
 		          ERR_error_string(ERR_peek_last_error(), NULL));
@@ -836,6 +979,7 @@ int run_tunnel(struct vpn_config *config)
 		.ssl_handle = NULL,
 		.ipv4.ns1_addr.s_addr = 0,
 		.ipv4.ns2_addr.s_addr = 0,
+		.ipv4.dns_suffix = NULL,
 		.on_ppp_if_up = on_ppp_if_up,
 		.on_ppp_if_down = on_ppp_if_down
 	};
