@@ -42,6 +42,9 @@ static char *uri_escape(const char *string)
 		if (allocated_len + 4 >= real_len) {
 			allocated_len += 16;
 			escaped = realloc(escaped, allocated_len);
+			// bail out if realloc fails
+			if (escaped == NULL)
+				return NULL;
 		}
 		if (isalnum(string[i]))
 			escaped[real_len++] = string[i];
@@ -96,6 +99,7 @@ static int pinentry_read(int from, char **retstr)
 					*retstr = strdup(strerror(errno));
 				return -1;
 			}
+			buf[bufsiz-1] = '\0';
 		}
 
 		ret = read(from, &buf[len], bufsiz - len);
@@ -113,7 +117,8 @@ static int pinentry_read(int from, char **retstr)
 		}
 		len += ret;
 	} while (buf[len - 1] != '\n');
-	*strchr(buf, '\n') = '\0';
+	// overwrite the newline with a null character
+	buf[len - 1] = '\0';
 
 	if (strcmp(buf, "OK") == 0 || strncmp(buf, "OK ", 3) == 0
 	    || strncmp(buf, "D ", 2) == 0) {
@@ -153,6 +158,7 @@ static int pinentry_exchange(int to, int from, char **retstr,
 	if (vdprintf(to, format, ap) == 0) {
 		if (retstr)
 			*retstr = strdup(strerror(errno));
+		va_end(ap);
 		return -1;
 	}
 	va_end(ap);
@@ -238,9 +244,13 @@ static void pinentry_read_password(const char *pinentry, const char *hint,
 		goto out;
 
 	escaped = uri_escape(hint);
-	pinentry_exchange(to_pinentry[1], from_pinentry[0], NULL,
-	                  "SETKEYINFO %s\n", escaped);
+	ret = pinentry_exchange(to_pinentry[1], from_pinentry[0], NULL,
+	                        "SETKEYINFO %s\n", escaped);
+	if (ret)
+		log_error("Failed to set keyinfo\n");
 	free(escaped);
+	if (ret)
+		goto out;
 
 	escaped = uri_escape(prompt);
 	ret = pinentry_exchange(to_pinentry[1], from_pinentry[0], &retstr,
