@@ -20,6 +20,7 @@
 
 #include <sys/types.h>
 
-void read_password(const char *prompt, char *pass, size_t len);
+void read_password(const char *pinentry, const char *hint,
+                   const char *prompt, char *pass, size_t len);
 
 #endif
